#define NOMINMAX
#include <windows.h>

#define DLL_EXPORT __declspec(dllexport)
#define MAX_BITMAP_SIZE 32768
#define BPP 4

#define for_iter(name, min, max) for (name = min; name < max; ++name)

/***********************************************************************************
* Structs
************************************************************************************/

typedef struct
{
    DWORD flags;
    DWORD klass;
    void (*dmark)(void*);
    void (*dfree)(void*);
    BYTE* data; // B = 0, G = 1, R = 2, A = 3
} RGSSCOLOR;

typedef struct
{
    DWORD unk1;
    DWORD unk2;
    BITMAPINFOHEADER* infoheader;
    RGSSCOLOR* firstRow;
    RGBQUAD* lastRow;
} RGSSBMINFO;

typedef struct
{
    DWORD unk1;
    DWORD unk2;
    RGSSBMINFO* bminfo;
} BITMAPSTRUCT;

typedef struct
{
    DWORD flags;
    DWORD klass;
    void (*dmark)(void*);
    void (*dfree)(void*);
    BITMAPSTRUCT* bm;
} RGSSBITMAP;

static float _srcXs[MAX_BITMAP_SIZE] = { 0.0f };
static int _x0s[MAX_BITMAP_SIZE] = { 0 };
static int _x1s[MAX_BITMAP_SIZE] = { 0 };
static float _rx0s[MAX_BITMAP_SIZE] = { 0.0f };
static float _rx1s[MAX_BITMAP_SIZE] = { 0.0f };
static float _srcYs[MAX_BITMAP_SIZE] = { 0.0f };
static int _y0s[MAX_BITMAP_SIZE] = { 0 };
static int _y1s[MAX_BITMAP_SIZE] = { 0 };
static float _ry0s[MAX_BITMAP_SIZE] = { 0.0f };
static float _ry1s[MAX_BITMAP_SIZE] = { 0.0f };

// returns 0 upon success
// not thread-safe!
int DLL_EXPORT LinearWrite(long destObject, int dx, int dy, int dw, int dh,
	long srcObject, int sx, int sy, int sw, int sh)
{
	RGSSBMINFO* destBitmap = ((RGSSBITMAP*)(destObject << 1))->bm->bminfo;
	RGSSBMINFO* srcBitmap = ((RGSSBITMAP*)(srcObject << 1))->bm->bminfo;
	if (destBitmap == NULL || srcBitmap == NULL)
	{
		return 1;
	}
	// used variables
	int destWidth = destBitmap->infoheader->biWidth;
	int destHeight = destBitmap->infoheader->biHeight;
	int srcWidth = srcBitmap->infoheader->biWidth;
	int srcHeight = srcBitmap->infoheader->biHeight;
	// correcting coordinates, aborting if something's wrong (e.g. uncorrectable out bounds)
	if (dx < 0)
	{
		dw += dx;
		dx = 0;
	}
	if (dw > destWidth - dx)
	{
		dw = destWidth - dx;
	}
	if (dw <= 0)
	{
		return 2;
	}
	if (dy < 0)
	{
		dh += dy;
		dy = 0;
	}
	if (dh > destHeight - dy)
	{
		dh = destHeight - dy;
	}
	if (dh <= 0)
	{
		return 2;
	}
	if (sx < 0)
	{
		sw += sx;
		sx = 0;
	}
	if (sw > srcWidth - sx)
	{
		sw = srcWidth - sx;
	}
	if (sw <= 0)
	{
		return 2;
	}
	if (sy < 0)
	{
		sh += sy;
		sy = 0;
	}
	if (sh > srcHeight - sy)
	{
		sh = srcHeight - sy;
	}
	if (sh <= 0)
	{
		return 2;
	}
	// more variables
	unsigned char* destData = (unsigned char*)destBitmap->firstRow;
	unsigned char* srcData = (unsigned char*)srcBitmap->firstRow;
	float fw = (sw - 1.0f) / dw;
	float fh = (sh - 1.0f) / dh;
	unsigned char* dest = NULL;
	unsigned char* ctl = NULL;
	unsigned char* ctr = NULL;
	unsigned char* cbl = NULL;
	unsigned char* cbr = NULL;
	int x = 0;
	int y = 0;
	// preparing all reused values
	for_iter (y, 0, dh)
	{
		_srcYs[y] = sy + y * fh;
		_y0s[y] = (int)_srcYs[y];
		_ry0s[y] = _srcYs[y] - _y0s[y];
		_y1s[y] = min(_y0s[y] + 1, sy + sh - 1);
		_ry1s[y] = 1.0f - _ry0s[y];
	}
	for_iter (x, 0, dw)
	{
		_srcXs[x] = sx + x * fw;
		_x0s[x] = (int)_srcXs[x];
		_rx0s[x] = _srcXs[x] - _x0s[x];
		_x1s[x] = min(_x0s[x] + 1, sx + sw - 1);
		_rx1s[x] = 1.0f - _rx0s[x];
	}
	for_iter (y, 0, dh)
	{
		for_iter (x, 0, dw)
		{
			dest = &destData[((dx + x) - (dy + y) * destWidth) * BPP];
			// linear interpolation
			ctl = &srcData[(_x0s[x] - _y0s[y] * srcWidth) * BPP];
			if (_rx0s[x] != 0.0f && _ry0s[y] != 0.0f)
			{
				ctr = &srcData[(_x1s[x] - _y0s[y] * srcWidth) * BPP];
				cbl = &srcData[(_x0s[x] - _y1s[y] * srcWidth) * BPP];
				cbr = &srcData[(_x1s[x] - _y1s[y] * srcWidth) * BPP];
				dest[2] = (unsigned char)(((ctl[2] * _ry1s[y] + cbl[2] * _ry0s[y]) * _rx1s[x] + (ctr[2] * _ry1s[y] + cbr[2] * _ry0s[y]) * _rx0s[x]));
				dest[1] = (unsigned char)(((ctl[1] * _ry1s[y] + cbl[1] * _ry0s[y]) * _rx1s[x] + (ctr[1] * _ry1s[y] + cbr[1] * _ry0s[y]) * _rx0s[x]));
				dest[0] = (unsigned char)(((ctl[0] * _ry1s[y] + cbl[0] * _ry0s[y]) * _rx1s[x] + (ctr[0] * _ry1s[y] + cbr[0] * _ry0s[y]) * _rx0s[x]));
				dest[3] = (unsigned char)(((ctl[3] * _ry1s[y] + cbl[3] * _ry0s[y]) * _rx1s[x] + (ctr[3] * _ry1s[y] + cbr[3] * _ry0s[y]) * _rx0s[x]));
			}
			else if (_rx0s[x] != 0.0f)
			{
				ctr = &srcData[(_x1s[x] - _y0s[y] * srcWidth) * BPP];
				dest[2] = (unsigned char)((ctl[2] * _rx1s[x] + ctr[2] * _rx0s[x]));
				dest[1] = (unsigned char)((ctl[1] * _rx1s[x] + ctr[1] * _rx0s[x]));
				dest[0] = (unsigned char)((ctl[0] * _rx1s[x] + ctr[0] * _rx0s[x]));
				dest[3] = (unsigned char)((ctl[3] * _rx1s[x] + ctr[3] * _rx0s[x]));
			}
			else if (_ry0s[y] != 0.0f)
			{
				cbl = &srcData[(_x0s[x] - _y1s[y] * srcWidth) * BPP];
				dest[2] = (unsigned char)((ctl[2] * _ry1s[y] + cbl[2] * _ry0s[y]));
				dest[1] = (unsigned char)((ctl[1] * _ry1s[y] + cbl[1] * _ry0s[y]));
				dest[0] = (unsigned char)((ctl[0] * _ry1s[y] + cbl[0] * _ry0s[y]));
				dest[3] = (unsigned char)((ctl[3] * _ry1s[y] + cbl[3] * _ry0s[y]));
			}
			else
			{
				memcpy(dest, ctl, 4);
			}
		}
	}
	return 0;
}
